class Scene(Entity):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Entity(parent=self, position=Vec3(-1.0334,0,3.0372), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-5.3896,0.002,9.6256), scale=Vec3(1,2,1), model='cube', origin=Vec3(0,-0.5,0), color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-2.1674,0.064,2.3704), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.3745,0.002,-6.6139), model='cube', color=color.hsv(209,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-0.2384,0.065,-0.8437), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(1.1839,0.003,8.2525), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.0499,0.067,7.5857), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(4.3956,0.004,-2.4644), model='cube', color=color.hsv(209,1,1), ignore=True)
        Entity(parent=self, position=Vec3(2.1659,0.067,-2.8756), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-1.4502,0.069,-6.4631), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-6.3671,0.001,4.0369), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-3.4524,0.001,3.39), scale=Vec3(1,2,1), model='cube', color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-7.5011,0.065,3.3701), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-4.7031,0.002,0.8228), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-5.8371,0.066,0.156), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-2.5641,0.003,2.7849), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-3.6981,0.067,2.1181), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-2.2987,0.004,-1.2091), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-3.4327,0.068,-1.8759), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-4.8313,0.005,-2.0694), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-5.9653,0.069,-2.7362), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-1.9876,-0.067,13.2403), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.9271,-0.067,12.5934), scale=Vec3(1,2,1), model='cube', color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-1.5359,-0.004,6.1062), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-0.3235,-0.066,10.0263), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.1282,-0.003,2.8922), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(1.8154,-0.065,11.9884), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.6814,-0.001,11.3216), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(2.0808,-0.064,7.9944), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.9468,0,7.3276), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(1.3989,-0.064,0.6668), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(3.1222,0.002,2.838), scale=Vec3(1,2,1), model='cube', color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(4.0105,0.004,2.2329), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(2.8765,0.068,1.5661), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.839,-0.066,7.2207), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(3.7537,-0.066,6.5738), scale=Vec3(1,2,1), model='cube', color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(2.5031,-0.065,4.0067), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(4.642,-0.064,5.9688), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(3.508,0,5.302), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(4.9074,-0.063,1.9748), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(3.7734,0.001,1.308), model='cube', color=color.hsv(330,1,1), ignore=True)
